import tqdm

import urllib.request
import subprocess
import sys

from spotdl.encode.encoders import EncoderFFmpeg
from spotdl.metadata.embedders import EmbedderDefault

import spotdl.util

CHUNK_SIZE = 16 * 1024

class Track:
    def __init__(self, metadata, cache_albumart=False):
        self.metadata = metadata
        self._chunksize = CHUNK_SIZE

        if cache_albumart:
            self._albumart_thread = self._cache_albumart()

        self._cache_albumart = cache_albumart

    def _cache_albumart(self):
        albumart_thread = spotdl.util.ThreadWithReturnValue(
            target=lambda url: urllib.request.urlopen(url).read(),
            args=(self.metadata["album"]["images"][0]["url"],)
        )
        albumart_thread.start()
        return albumart_thread

    def calculate_total_chunks(self, filesize):
        return (filesize // self._chunksize) + 1

    def make_progress_bar(self, total_chunks):
        progress_bar = tqdm.trange(
            total_chunks,
            unit_scale=(self._chunksize // 1024),
            unit="KiB",
            dynamic_ncols=True,
            bar_format='{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt}KiB '
                '[{elapsed}<{remaining}, {rate_fmt}{postfix}]',
        )
        return progress_bar

    def download_while_re_encoding(self, stream, target_path, target_encoding=None,
                                   encoder=EncoderFFmpeg(must_exist=False), show_progress=True):
        total_chunks = self.calculate_total_chunks(stream["filesize"])
        process = encoder.re_encode_from_stdin(
            stream["encoding"],
            target_path,
            target_encoding=target_encoding
        )
        response = stream["connection"]

        progress_bar = self.make_progress_bar(total_chunks)
        for _ in progress_bar:
            chunk = response.read(self._chunksize)
            process.stdin.write(chunk)

        process.stdin.close()
        process.wait()

    def download(self, stream, target_path, show_progress=True):
        total_chunks = self.calculate_total_chunks(stream["filesize"])
        progress_bar = self.make_progress_bar(total_chunks)
        response = stream["connection"]

        def writer(response, progress_bar, file_io):
            for _ in progress_bar:
                chunk = response.read(self._chunksize)
                file_io.write(chunk)

        write_to_stdout = target_path == "-"
        if write_to_stdout:
            file_io = sys.stdout.buffer
            writer(response, progress_bar, file_io)
        else:
            with open(target_path, "wb") as file_io:
                writer(response, progress_bar, file_io)

    def re_encode(self, input_path, target_path, target_encoding=None,
                  encoder=EncoderFFmpeg(must_exist=False), show_progress=True):
        stream = self.metadata["streams"].getbest()
        total_chunks = self.calculate_total_chunks(stream["filesize"])
        process = encoder.re_encode_from_stdin(
            stream["encoding"],
            target_path,
            target_encoding=target_encoding
        )
        with open(input_path, "rb") as fin:
            for _ in tqdm.trange(total_chunks):
                chunk = fin.read(self._chunksize)
                process.stdin.write(chunk)

        process.stdin.close()
        process.wait()

    def apply_metadata(self, input_path, encoding=None, embedder=EmbedderDefault()):
        if self._cache_albumart:
            albumart = self._albumart_thread.join()
        else:
            albumart = None

        embedder.apply_metadata(
            input_path,
            self.metadata,
            cached_albumart=albumart,
            encoding=encoding,
        )




import appdirs
import yaml
import os

import spotdl.util
import logging
logger = logging.getLogger(__name__)

DEFAULT_CONFIGURATION = {
    "spotify-downloader": {
        "manual": False,
        "no_metadata": False,
        "no_encode": False,
        "overwrite": "prompt",
        "quality": "best",
        "input_ext": "automatic",
        "output_ext": "mp3",
        "write_to": None,
        "trim_silence": False,
        "search_format": "{artist} - {track-name} lyrics",
        "dry_run": False,
        "no_spaces": False,
        # "processor": "synchronous",
        "output_file": "{artist} - {track-name}.{output-ext}",
        "skip_file": None,
        "write_successful_file": None,
        "spotify_client_id": "4fe3fecfe5334023a1472516cc99d805",
        "spotify_client_secret": "0f02b7c483c04257984695007a4a8d5c",
        "log_level": "INFO",
    }
}

DEFAULT_CONFIG_FILE = os.path.join(
    appdirs.user_config_dir(),
    "spotdl",
    "config.yml"
)

def read_config(config_file):
    with open(config_file, "r") as ymlfile:
        config = yaml.safe_load(ymlfile)
    return config


def dump_config(config_file=None, config=DEFAULT_CONFIGURATION):
    if config_file is None:
        config = yaml.dump(config, default_flow_style=False)
        return config

    with open(config_file, "w") as ymlfile:
        yaml.dump(config, ymlfile, default_flow_style=False)




class NoYouTubeVideoFoundError(Exception):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)


class NoYouTubeVideoMatchError(Exception):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)


class ArgumentError(Exception):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)




class MetadataNotFoundError(Exception):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)


class SpotifyMetadataNotFoundError(MetadataNotFoundError):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)


class YouTubeMetadataNotFoundError(MetadataNotFoundError):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)




def format_string(string, metadata, output_extension="", sanitizer=lambda s: s):
    formats = {
        "{track-name}"   : metadata["name"],
        "{artist}"       : metadata["artists"][0]["name"],
        "{album}"        : metadata["album"]["name"],
        "{album-artist}" : metadata["artists"][0]["name"],
        "{genre}"        : metadata["genre"],
        "{disc-number}"  : metadata["disc_number"],
        "{duration}"     : metadata["duration"],
        "{year}"         : metadata["year"],
        "{original-date}": metadata["release_date"],
        "{track-number}" : str(metadata["track_number"]).zfill(len(str(metadata["total_tracks"]))),
        "{total-tracks}" : metadata["total_tracks"],
        "{isrc}"         : metadata["external_ids"]["isrc"],
        "{track-id}"     : metadata.get("id", ""),
        "{output-ext}"   : output_extension,
    }

    for key, value in formats.items():
        string = string.replace(key, sanitizer(str(value)))

    return string




from spotdl.authorize.authorize_base import AuthorizeBase

from spotdl.authorize.exceptions import AuthorizationError
from spotdl.authorize.exceptions import SpotifyAuthorizationError
from spotdl.authorize.exceptions import YouTubeAuthorizationError




import subprocess
import os

from spotdl.encode import EncoderBase
from spotdl.encode.exceptions import EncoderNotFoundError
from spotdl.encode.exceptions import FFmpegNotFoundError

import logging
logger = logging.getLogger(__name__)


# Key: from format
# Subkey: to format
RULES = {
    "m4a": {
        "mp3": "-codec:v copy -codec:a libmp3lame -ar 48000",
        "opus": "-codec:a libopus -vbr on",
        "m4a": "-acodec copy",
        "flac": "-codec:a flac -ar 48000",
    },
    "opus": {
        "mp3": "-codec:a libmp3lame -ar 48000",
        "m4a": "-cutoff 20000 -codec:a aac -ar 48000",
        "flac": "-codec:a flac -ar 48000",
    },
}


class EncoderFFmpeg(EncoderBase):
    def __init__(self, encoder_path="ffmpeg", must_exist=True):
        _loglevel = "-hide_banner -nostats -v panic"
        _additional_arguments = ["-b:a", "192k", "-vn"]
        try:
            super().__init__(encoder_path, must_exist, _loglevel, _additional_arguments)
        except EncoderNotFoundError as e:
            raise FFmpegNotFoundError(e.args[0])
        self._rules = RULES

    def set_trim_silence(self):
        self.set_argument("-af silenceremove=start_periods=1")

    def get_encoding(self, path):
        return super().get_encoding(path)

    def _generate_encoding_arguments(self, input_encoding, target_encoding):
        initial_arguments = self._rules.get(input_encoding)
        if initial_arguments is None:
            raise TypeError(
                'The input format ("{}") is not supported.'.format(
                input_encoding,
            ))
        arguments = initial_arguments.get(target_encoding)
        if arguments is None:
            raise TypeError(
                'The output format ("{}") is not supported.'.format(
                target_encoding,
            ))
        return arguments

    def set_debuglog(self):
        self._loglevel = "-loglevel debug"

    def _generate_encode_command(self, input_path, target_file,
                                 input_encoding=None, target_encoding=None):
        if input_encoding is None:
            input_encoding = self.get_encoding(input_path)
        if target_encoding is None:
            target_encoding = self.get_encoding(target_file)
        arguments = self._generate_encoding_arguments(
            input_encoding,
            target_encoding
        )
        command = [self.encoder_path] \
            + ["-y", "-nostdin"] \
            + self._loglevel.split() \
            + ["-i", input_path] \
            + arguments.split() \
            + self._additional_arguments \
            + ["-f", self.target_format_from_encoding(target_encoding)] \
            + [target_file]

        return command

    def re_encode(self, input_path, target_file, target_encoding=None, delete_original=False):
        encode_command = self._generate_encode_command(
            input_path,
            target_file,
            target_encoding=target_encoding
        )
        logger.debug("Calling FFmpeg with:\n{command}".format(
            command=encode_command,
        ))
        process = subprocess.Popen(encode_command)
        process.wait()
        encode_successful = process.returncode == 0
        if encode_successful and delete_original:
            os.remove(input_path)
        return process

    def re_encode_from_stdin(self, input_encoding, target_file, target_encoding=None):
        encode_command = self._generate_encode_command(
            "-",
            target_file,
            input_encoding=input_encoding,
            target_encoding=target_encoding,
        )
        logger.debug("Calling FFmpeg with:\n{command}".format(
            command=encode_command,
        ))
        process = subprocess.Popen(encode_command, stdin=subprocess.PIPE)
        return process




from spotdl.metadata.providers import ProviderSpotify
from spotdl.metadata.providers import ProviderYouTube
from spotdl.lyrics.providers import Genius
from spotdl.lyrics.exceptions import LyricsNotFoundError

import spotdl.metadata
import spotdl.util
from spotdl.metadata.exceptions import SpotifyMetadataNotFoundError

from spotdl.command_line.exceptions import NoYouTubeVideoFoundError
from spotdl.command_line.exceptions import NoYouTubeVideoMatchError

import sys
import logging
logger = logging.getLogger(__name__)

PROVIDERS = {
    "spotify": ProviderSpotify,
    "youtube": ProviderYouTube,
}


def prompt_for_youtube_search_result(videos):
    max_index_length = len(str(len(videos)))
    max_title_length = max(len(v["title"]) for v in videos)
    msg = "{index:>{max_index}}. Skip downloading this track".format(
        index=0,
        max_index=max_index_length,
    )
    print(msg, file=sys.stderr)
    for index, video in enumerate(videos, 1):
        vid_details = "{index:>{max_index}}. {title:<{max_title}}\n{new_line_gap}  {url} [{duration}]".format(
            index=index,
            max_index=max_index_length,
            title=video["title"],
            max_title=max_title_length,
            new_line_gap=" " * max_index_length,
            url=video["url"],
            duration=video["duration"],
        )
        print(vid_details, file=sys.stderr)
    print("", file=sys.stderr)

    selection = spotdl.util.prompt_user_for_selection(range(1, len(videos)+1))

    if selection is None:
        return None
    return videos[selection-1]


class MetadataSearch:
    def __init__(self, track, lyrics=False, yt_search_format="{artist} - {track-name}", yt_manual=False, providers=PROVIDERS):
        self.track = track
        self.track_type = spotdl.util.track_type(track)
        self.lyrics = lyrics
        self.yt_search_format = yt_search_format
        self.yt_manual = yt_manual
        self.providers = {}
        for provider, parent in providers.items():
            self.providers[provider] = parent()
        self.lyric_provider = Genius()

    def get_lyrics(self, query):
        try:
            lyrics = self.lyric_provider.from_query(query)
        except LyricsNotFoundError as e:
            logger.warning(e.args[0])
            lyrics = None
        return lyrics

    def _make_lyric_search_query(self, metadata):
        if self.track_type == "query":
            lyric_query = self.track
        else:
            lyric_search_format = "{artist} - {track-name}"
            lyric_query = spotdl.metadata.format_string(
                lyric_search_format,
                metadata
            )
        return lyric_query

    def on_youtube_and_spotify(self):
        track_type_mapper = {
            "spotify": self._on_youtube_and_spotify_for_type_spotify,
            "youtube": self._on_youtube_and_spotify_for_type_youtube,
            "query": self._on_youtube_and_spotify_for_type_query,
        }
        caller = track_type_mapper[self.track_type]
        metadata = caller()

        if not self.lyrics:
            return metadata

        lyric_query = self._make_lyric_search_query(metadata)
        metadata["lyrics"] = spotdl.util.ThreadWithReturnValue(
            target=self.get_lyrics,
            args=(lyric_query,),
        )

        return metadata

    def on_youtube(self):
        track_type_mapper = {
            "spotify": self._on_youtube_for_type_spotify,
            "youtube": self._on_youtube_for_type_youtube,
            "query": self._on_youtube_for_type_query,
        }
        caller = track_type_mapper[self.track_type]
        metadata = caller(self.track)

        if not self.lyrics:
            return metadata

        lyric_query = self._make_lyric_search_query(metadata)
        metadata["lyrics"] = spotdl.util.ThreadWithReturnValue(
            target=self.get_lyrics,
            arguments=(lyric_query,),
        )

        return metadata

    def on_spotify(self):
        track_type_mapper = {
            "spotify": self._on_spotify_for_type_spotify,
            "youtube": self._on_spotify_for_type_youtube,
            "query": self._on_spotify_for_type_query,
        }
        caller = track_type_mapper[self.track_type]
        metadata = caller(self.track)

        if not self.lyrics:
            return metadata

        lyric_query = self._make_lyric_search_query(metadata)
        metadata["lyrics"] = spotdl.util.ThreadWithReturnValue(
            target=self.get_lyrics,
            arguments=(lyric_query,),
        )

        return metadata

    def best_on_youtube_search(self):
        track_type_mapper = {
            "spotify": self._best_on_youtube_search_for_type_spotify,
            "youtube": self._best_on_youtube_search_for_type_youtube,
            "query": self._best_on_youtube_search_for_type_query,
        }
        caller = track_type_mapper[self.track_type]
        video = caller(self.track)
        return video

    def _best_on_youtube_search_for_type_query(self, query):
        videos = self.providers["youtube"].search(query)
        if not videos:
            raise NoYouTubeVideoFoundError(
                'YouTube returned no videos for the search query "{}".'.format(query)
            )
        if self.yt_manual:
            video = prompt_for_youtube_search_result(videos)
        else:
            video = videos.bestmatch()

        if video is None:
            raise NoYouTubeVideoMatchError(
                'No matching videos found on YouTube for the search query "{}".'.format(
                    query
                )
            )
        return video

    def _best_on_youtube_search_for_type_youtube(self, url):
        video = self._best_on_youtube_search_for_type_query(url)
        return video

    def _best_on_youtube_search_for_type_spotify(self, url):
        spotify_metadata = self._on_spotify_for_type_spotify(self.track)
        search_query = spotdl.metadata.format_string(self.yt_search_format, spotify_metadata)
        video = self._best_on_youtube_search_for_type_query(search_query)
        return video

    def _on_youtube_and_spotify_for_type_spotify(self):
        logger.debug("Extracting YouTube and Spotify metadata for input Spotify URI.")
        spotify_metadata = self._on_spotify_for_type_spotify(self.track)
        search_query = spotdl.metadata.format_string(self.yt_search_format, spotify_metadata)
        youtube_video = self._best_on_youtube_search_for_type_spotify(search_query)
        youtube_metadata = self.providers["youtube"].from_url(youtube_video["url"])
        metadata = spotdl.util.merge_copy(
            youtube_metadata,
            spotify_metadata
        )
        return metadata

    def _on_youtube_and_spotify_for_type_youtube(self):
        logger.debug("Extracting YouTube and Spotify metadata for input YouTube URL.")
        youtube_metadata = self._on_youtube_for_type_youtube(self.track)
        search_query = spotdl.metadata.format_string("{track-name}", youtube_metadata)
        spotify_metadata = self._on_spotify_for_type_query(search_query)
        metadata = spotdl.util.merge_copy(
            youtube_metadata,
            spotify_metadata
        )
        return metadata

    def _on_youtube_and_spotify_for_type_query(self):
        logger.debug("Extracting YouTube and Spotify metadata for input track query.")
        search_query = self.track
        # Make use of threads here to search on both YouTube & Spotify
        # at the same time.
        spotify_metadata = spotdl.util.ThreadWithReturnValue(
            target=self._on_spotify_for_type_query,
            args=(search_query,)
        )
        spotify_metadata.start()
        youtube_metadata = self._on_youtube_for_type_query(search_query)
        metadata = spotdl.util.merge_copy(
            youtube_metadata,
            spotify_metadata.join()
        )
        return metadata

    def _on_youtube_for_type_spotify(self):
        logger.debug("Extracting YouTube metadata for input Spotify URI.")
        spotify_metadata = self._on_spotify_for_type_spotify(self.track)
        search_query = spotdl.metadata.format_string(self.yt_search_format, spotify_metadata)
        youtube_video = self._best_on_youtube_search_for_type_spotify(search_query)
        youtube_metadata = self.providers["youtube"].from_url(youtube_video["url"])
        return youtube_metadata

    def _on_youtube_for_type_youtube(self, url):
        logger.debug("Extracting YouTube metadata for input YouTube URL.")
        youtube_metadata = self.providers["youtube"].from_url(url)
        return youtube_metadata

    def _on_youtube_for_type_query(self, query):
        logger.debug("Extracting YouTube metadata for input track query.")
        youtube_video = self._best_on_youtube_search_for_type_query(query)
        youtube_metadata = self.providers["youtube"].from_url(youtube_video["url"])
        return youtube_metadata

    def _on_spotify_for_type_youtube(self, url):
        logger.debug("Extracting Spotify metadata for input YouTube URL.")
        youtube_metadata = self.providers["youtube"].from_url(url)
        search_query = spotdl.metadata.format_string("{track-name}", youtube_metadata)
        spotify_metadata = self.providers["spotify"].from_query(search_query)
        return spotify_metadata

    def _on_spotify_for_type_spotify(self, url):
        logger.debug("Extracting Spotify metadata for input Spotify URI.")
        spotify_metadata = self.providers["spotify"].from_url(url)
        return spotify_metadata

    def _on_spotify_for_type_query(self, query):
        logger.debug("Extracting Spotify metadata for input track query.")
        try:
            spotify_metadata = self.providers["spotify"].from_query(query)
        except SpotifyMetadataNotFoundError as e:
            logger.warn(e.args[0])
            spotify_metadata = {}
        return spotify_metadata




from abc import ABC
from abc import abstractmethod

class StreamsBase(ABC):
    @abstractmethod
    def __init__(self, streams):
        """
        This method must parse audio streams into a list of
        dictionaries with the keys:
        "bitrate", "download_url", "encoding", "filesize".

        The list should typically be sorted in descending order
        based on the audio stream's bitrate.

        This sorted list must be assigned to `self.all`.
        """
        self.all = streams

    def getbest(self):
        """
        This method must return the audio stream with the
        highest bitrate.
        """
        return self.all[0]

    def getworst(self):
        """
        This method must return the audio stream with the
        lowest bitrate.
        """
        return self.all[-1]


class ProviderBase(ABC):
    def set_credentials(self, client_id, client_secret):
        """
        This method may or not be used depending on
        whether the metadata provider requires authentication
        or not.
        """
        pass

    @abstractmethod
    def from_url(self, url):
        """
        This method must return track metadata from the
        corresponding Spotify URL.
        """
        pass

    def from_query(self, query):
        """
        This method must return track metadata from the
        corresponding search query.
        """
        raise NotImplementedError

    @abstractmethod
    def metadata_to_standard_form(self, metadata):
        """
        This method must transform the fetched metadata
        into a format consistent with all other metadata
        providers, for easy utilization.
        """
        pass




from mutagen.easyid3 import EasyID3
from mutagen.id3 import ID3, TORY, TYER, TPUB, APIC, USLT, COMM
from mutagen.mp4 import MP4, MP4Cover
from mutagen.flac import Picture, FLAC

import urllib.request

from spotdl.metadata import EmbedderBase

import logging
logger = logging.getLogger(__name__)

# Apple has specific tags - see mutagen docs -
# http://mutagen.readthedocs.io/en/latest/api/mp4.html
M4A_TAG_PRESET = {
    "album": "\xa9alb",
    "artist": "\xa9ART",
    "date": "\xa9day",
    "title": "\xa9nam",
    "year": "\xa9day",
    "originaldate": "purd",
    "comment": "\xa9cmt",
    "group": "\xa9grp",
    "writer": "\xa9wrt",
    "genre": "\xa9gen",
    "tracknumber": "trkn",
    "albumartist": "aART",
    "discnumber": "disk",
    "cpil": "cpil",
    "albumart": "covr",
    "copyright": "cprt",
    "tempo": "tmpo",
    "lyrics": "\xa9lyr",
    "comment": "\xa9cmt",
}

TAG_PRESET = {}
for key in M4A_TAG_PRESET.keys():
    TAG_PRESET[key] = key


class EmbedderDefault(EmbedderBase):
    supported_formats = ("mp3", "m4a", "flac")

    def __init__(self):
        super().__init__()
        self._m4a_tag_preset = M4A_TAG_PRESET
        self._tag_preset = TAG_PRESET
        # self.provider = "spotify" if metadata["spotify_metadata"] else "youtube"

    def as_mp3(self, path, metadata, cached_albumart=None):
        """ Embed metadata to MP3 files. """
        logger.debug('Writing MP3 metadata to "{path}".'.format(path=path))
        # EasyID3 is fun to use ;)
        # For supported easyid3 tags:
        # https://github.com/quodlibet/mutagen/blob/master/mutagen/easyid3.py
        # Check out somewhere at end of above linked file
        audiofile = EasyID3(path)
        self._embed_basic_metadata(audiofile, metadata, "mp3", preset=TAG_PRESET)
        audiofile["media"] = metadata["type"]
        audiofile["author"] = metadata["artists"][0]["name"]
        audiofile["lyricist"] = metadata["artists"][0]["name"]
        audiofile["arranger"] = metadata["artists"][0]["name"]
        audiofile["performer"] = metadata["artists"][0]["name"]
        provider = metadata["provider"]
        audiofile["website"] = metadata["external_urls"][provider]
        audiofile["length"] = str(metadata["duration"])
        if metadata["publisher"]:
            audiofile["encodedby"] = metadata["publisher"]
        if metadata["external_ids"]["isrc"]:
            audiofile["isrc"] = metadata["external_ids"]["isrc"]
        audiofile.save(v2_version=3)

        # For supported id3 tags:
        # https://github.com/quodlibet/mutagen/blob/master/mutagen/id3/_frames.py
        # Each class represents an id3 tag
        audiofile = ID3(path)
        if metadata["year"]:
            audiofile["TORY"] = TORY(encoding=3, text=metadata["year"])
            audiofile["TYER"] = TYER(encoding=3, text=metadata["year"])
        if metadata["publisher"]:
            audiofile["TPUB"] = TPUB(encoding=3, text=metadata["publisher"])
        provider = metadata["provider"]
        audiofile["COMM"] = COMM(
            encoding=3, text=metadata["external_urls"][provider]
        )
        if metadata["lyrics"]:
            audiofile["USLT"] = USLT(
                encoding=3, desc=u"Lyrics", text=metadata["lyrics"]
            )
        if cached_albumart is None:
            cached_albumart = urllib.request.urlopen(
                metadata["album"]["images"][0]["url"]
            ).read()
            albumart.close()
        try:
            audiofile["APIC"] = APIC(
                encoding=3,
                mime="image/jpeg",
                type=3,
                desc=u"Cover",
                data=cached_albumart,
            )
        except IndexError:
            pass

        audiofile.save(v2_version=3)

    def as_m4a(self, path, metadata, cached_albumart=None):
        """ Embed metadata to M4A files. """
        logger.debug('Writing M4A metadata to "{path}".'.format(path=path))
        audiofile = MP4(path)
        self._embed_basic_metadata(audiofile, metadata, "m4a", preset=M4A_TAG_PRESET)
        if metadata["year"]:
            audiofile[M4A_TAG_PRESET["year"]] = metadata["year"]
        provider = metadata["provider"]
        audiofile[M4A_TAG_PRESET["comment"]] = metadata["external_urls"][provider]
        if metadata["lyrics"]:
            audiofile[M4A_TAG_PRESET["lyrics"]] = metadata["lyrics"]
        try:
            if cached_albumart is None:
                cached_albumart = urllib.request.urlopen(
                    metadata["album"]["images"][0]["url"]
                ).read()
                albumart.close()
            audiofile[M4A_TAG_PRESET["albumart"]] = [
                MP4Cover(cached_albumart, imageformat=MP4Cover.FORMAT_JPEG)
            ]
        except IndexError:
            pass

        audiofile.save()

    def as_flac(self, path, metadata, cached_albumart=None):
        logger.debug('Writing FLAC metadata to "{path}".'.format(path=path))
        audiofile = FLAC(path)
        self._embed_basic_metadata(audiofile, metadata, "flac")
        if metadata["year"]:
            audiofile["year"] = metadata["year"]
        provider = metadata["provider"]
        audiofile["comment"] = metadata["external_urls"][provider]
        if metadata["lyrics"]:
            audiofile["lyrics"] = metadata["lyrics"]

        image = Picture()
        image.type = 3
        image.desc = "Cover"
        image.mime = "image/jpeg"
        if cached_albumart is None:
            cached_albumart = urllib.request.urlopen(
                metadata["album"]["images"][0]["url"]
            ).read()
            albumart.close()
        image.data = cached_albumart
        audiofile.add_picture(image)

        audiofile.save()

    def _embed_basic_metadata(self, audiofile, metadata, encoding, preset=TAG_PRESET):
        audiofile[preset["artist"]] = metadata["artists"][0]["name"]
        if metadata["album"]["artists"][0]["name"]:
            audiofile[preset["albumartist"]] = metadata["album"]["artists"][0]["name"]
        if metadata["album"]["name"]:
            audiofile[preset["album"]] = metadata["album"]["name"]
        audiofile[preset["title"]] = metadata["name"]
        if metadata["release_date"]:
            audiofile[preset["date"]] = metadata["release_date"]
            audiofile[preset["originaldate"]] = metadata["release_date"]
        if metadata["genre"]:
            audiofile[preset["genre"]] = metadata["genre"]
        if metadata["copyright"]:
            audiofile[preset["copyright"]] = metadata["copyright"]
        if encoding == "flac":
            audiofile[preset["discnumber"]] = str(metadata["disc_number"])
        else:
            audiofile[preset["discnumber"]] = [(metadata["disc_number"], 0)]
        zfilled_track_number = str(metadata["track_number"]).zfill(len(str(metadata["total_tracks"])))
        if encoding == "flac":
            audiofile[preset["tracknumber"]] = zfilled_track_number
        else:
            if preset["tracknumber"] == TAG_PRESET["tracknumber"]:
                audiofile[preset["tracknumber"]] = "{}/{}".format(
                    zfilled_track_number, metadata["total_tracks"]
                )
            else:
                audiofile[preset["tracknumber"]] = [
                    (metadata["track_number"], metadata["total_tracks"])
                ]




from abc import ABC
from abc import abstractmethod

class AuthorizeBase(ABC):
    """
    Defined service authenticators must inherit from this abstract
    base class and implement their own functionality for the below
    defined methods.
    """

    @abstractmethod
    def authorize(self):
        """
        This method must authorize with the corresponding service
        and return an object that can be utilized in making
        authenticated requests.
        """
        pass




from bs4 import BeautifulSoup
import urllib.request
import json

from spotdl.lyrics.lyric_base import LyricBase
from spotdl.lyrics.exceptions import LyricsNotFoundError

import logging
logger = logging.getLogger(__name__)

BASE_URL = "https://genius.com"
BASE_SEARCH_URL = BASE_URL + "/api/search/multi?per_page=1&q="

# FIXME: Make Genius a metadata provider instead of lyric provider
#        Since, Genius parses additional metadata too (such as track
#        name, artist name, albumart url). For example, fetch this URL:
#        https://genius.com/api/search/multi?per_page=1&q=artist+trackname

class Genius(LyricBase):
    def __init__(self):
        self.base_url = BASE_URL
        self.base_search_url = BASE_SEARCH_URL

    def guess_lyric_url_from_artist_and_track(self, artist, track):
        """
        Returns the possible lyric URL for the track available on
        Genius. This may not always be a valid URL.
        """
        query = "/{} {} lyrics".format(artist, track)
        query = query.replace(" ", "-")
        encoded_query = urllib.request.quote(query)
        lyric_url = self.base_url + encoded_query
        return lyric_url

    def _fetch_url_page(self, url, timeout=None):
        """
        Makes a GET request to the given lyrics page URL and returns
        the HTML content in the case of a valid response.
        """
        request = urllib.request.Request(url)
        request.add_header("User-Agent", "urllib")
        try:
            response = urllib.request.urlopen(request, timeout=timeout)
        except urllib.request.HTTPError:
            raise LyricsNotFoundError(
                "Could not find Genius lyrics at URL: {}".format(url)
            )
        else:
            return response.read()

    def _get_lyrics_text(self, paragraph):
        """
        Extracts and returns the lyric content from the provided HTML.
        """
        if paragraph:
            return paragraph.get_text()
        else:
            raise LyricsNotFoundError(
                "The lyrics for this track are yet to be released on Genius."
            )

    def _fetch_search_page(self, url, timeout=None):
        """
        Returns search results from a given URL in JSON.
        """
        request = urllib.request.Request(url)
        request.add_header("User-Agent", "urllib")
        response = urllib.request.urlopen(request, timeout=timeout)
        metadata = json.loads(response.read())
        if len(metadata["response"]["sections"][0]["hits"]) == 0:
            raise LyricsNotFoundError(
                "Genius returned no lyric results for the search URL: {}".format(url)
            )
        return metadata

    def best_matching_lyric_url_from_query(self, query):
        """
        Returns the best matching track's URL from a given query.
        """
        encoded_query = urllib.request.quote(query.replace(" ", "+"))
        search_url = self.base_search_url + encoded_query
        logger.debug('Fetching Genius search results from "{}".'.format(search_url))
        metadata = self._fetch_search_page(search_url)

        lyric_url = None
        for section in metadata["response"]["sections"]:
            result = section["hits"][0]["result"]
            try:
                lyric_url = result["path"]
                break
            except KeyError:
                pass

        if lyric_url is None:
            raise LyricsNotFoundError(
                "Could not find any valid lyric paths in Genius "
                "lyrics API response for the query {}.".format(query)
            )

        return self.base_url + lyric_url

    def from_query(self, query, linesep="\n", timeout=None):
        """
        Returns the lyric string for the track best matching the
        given query.
        """
        logger.debug('Fetching lyrics for the search query on "{}".'.format(query))
        try:
            lyric_url = self.best_matching_lyric_url_from_query(query)
        except LyricsNotFoundError:
            raise LyricsNotFoundError(
                'Genius returned no lyric results for the search query "{}".'.format(query)
            )
        else:
            return self.from_url(lyric_url, linesep, timeout=timeout)

    def from_artist_and_track(self, artist, track, linesep="\n", timeout=None):
        """
        Returns the lyric string for the given artist and track
        by making scraping search results and fetching the first
        result.
        """
        lyric_url = self.guess_lyric_url_from_artist_and_track(artist, track)
        return self.from_url(lyric_url, linesep, timeout=timeout)

    def from_url(self, url, linesep="\n", retries=5, timeout=None):
        """
        Returns the lyric string for the given URL.
        """
        logger.debug('Fetching lyric text from "{}".'.format(url))
        lyric_html_page = self._fetch_url_page(url, timeout=timeout)
        soup = BeautifulSoup(lyric_html_page, "html.parser")
        paragraph = soup.find("p")
        # If <p> has a class (like <p class="bla">), then we got an invalid
        # response. Retry in such a case.
        invalid_response = paragraph.get("class") is not None
        to_retry = retries > 0 and invalid_response
        if to_retry:
            logger.debug(
                "Retrying since Genius returned invalid response for search "
                "results. Retries left: {retries}.".format(retries=retries)
            )
            return self.from_url(url, linesep=linesep, retries=retries-1, timeout=timeout)

        if invalid_response:
            raise LyricsNotFoundError(
                'Genius returned invalid response for the search URL "{}".'.format(url)
            )
        lyrics = self._get_lyrics_text(paragraph)
        return lyrics.replace("\n", linesep)




import os
import sys
import math
import urllib.request
import threading

import logging
logger = logging.getLogger(__name__)


try:
    import winreg
except ImportError:
    pass

try:
    from slugify import SLUG_OK, slugify
except ImportError:
    logger.error("Oops! `unicode-slugify` was not found.")
    logger.info("Please remove any other slugify library and install `unicode-slugify`")
    raise


# This has been referred from
# https://stackoverflow.com/a/6894023/6554943
# It's because threaded functions do not return by default
# Whereas this will return the value when `join` method
# is called.
class ThreadWithReturnValue(threading.Thread):
    def __init__(self, target=lambda: None, args=()):
        super().__init__(target=target, args=args)
        self._return = None

    def run(self):
        if self._target is not None:
            self._return = self._target(
                *self._args,
                **self._kwargs
            )

    def join(self, *args, **kwargs):
        super().join(*args, **kwargs)
        return self._return


def merge_copy(base, overrider):
    return merge(base.copy(), overrider)

def merge(base, overrider):
    """ Override base dict with an overrider dict, recursively. """
    for key, value in overrider.items():
        if isinstance(value, dict):
            subitem = base.setdefault(key, {})
            merge(subitem, value)
        else:
            base[key] = value

    return base


def prompt_user_for_selection(items):
    """ Let the user input a choice. """
    logger.info("Enter a number:")
    while True:
        try:
            the_chosen_one = int(input("> "))
            if 1 <= the_chosen_one <= len(items):
                return items[the_chosen_one - 1]
            elif the_chosen_one == 0:
                return None
            else:
                logger.warning("Choose a valid number!")
        except ValueError:
            logger.warning("Choose a valid number!")


def is_spotify(track):
    """ Check if the input song is a Spotify link. """
    status = len(track) == 22 and track.replace(" ", "%20") == track
    status = status or track.find("spotify") > -1
    return status


def is_youtube(track):
    """ Check if the input song is a YouTube link. """
    status = len(track) == 11 and track.replace(" ", "%20") == track
    status = status and not track.lower() == track
    status = status or "youtube.com/watch?v=" in track
    return status


def track_type(track):
    track_types = {
        "spotify": is_spotify,
        "youtube": is_youtube,
    }
    for provider, fn in track_types.items():
        if fn(track):
            return provider
    return "query"


def sanitize(string, ok="&-_()[]{}", spaces_to_underscores=False):
    """ Generate filename of the song to be downloaded. """
    if spaces_to_underscores:
        string = string.replace(" ", "_")
    # replace slashes with "-" to avoid directory creation errors
    string = string.replace("/", "-").replace("\\", "-")
    # slugify removes any special characters
    string = slugify(string, ok=ok, lower=False, spaces=True)
    return string


def get_sec(time_str):
    if ":" in time_str:
        splitter = ":"
    elif "." in time_str:
        splitter = "."
    else:
        raise ValueError(
            "No expected character found in {} to split" "time values.".format(time_str)
        )
    v = time_str.split(splitter, 3)
    v.reverse()
    sec = 0
    if len(v) > 0:  # seconds
        sec += int(v[0])
    if len(v) > 1:  # minutes
        sec += int(v[1]) * 60
    if len(v) > 2:  # hours
        sec += int(v[2]) * 3600
    return sec


def remove_duplicates(elements, condition=lambda _: True, operation=lambda x: x):
    """
    Removes duplicates from a list whilst preserving order.

    We could directly call `set()` on the list but it changes
    the order of elements.
    """

    local_set = set()
    local_set_add = local_set.add
    filtered_list = []
    for x in elements:
        if condition(x) and not (x in local_set or local_set_add(x)):
            operated = operation(x)
            filtered_list.append(operated)
            local_set_add(operated)
    return filtered_list


def titlecase(string):
    return " ".join(word.capitalize() for word in string.split())


def readlines_from_nonbinary_file(path):
    with open(path, "r") as fin:
        lines = fin.read().splitlines()
    return lines


def writelines_to_nonbinary_file(path, lines):
    with open(path, "w") as fout:
        fout.writelines(map(lambda x: x + "\n", lines))




import spotipy
import spotipy.oauth2 as oauth2

from spotdl.metadata import ProviderBase
from spotdl.metadata.exceptions import SpotifyMetadataNotFoundError

from spotdl.authorize.services import AuthorizeSpotify
import spotdl.util

import logging
logger = logging.getLogger(__name__)


class ProviderSpotify(ProviderBase):
    def __init__(self, spotify=None):
        if spotify is None:
            spotify = AuthorizeSpotify()
        self.spotify = spotify

    def set_credentials(self, client_id, client_secret):
        token = self._generate_token(client_id, client_secret)
        self.spotify = spotipy.Spotify(auth=token)

    def from_url(self, url):
        logger.debug('Fetching Spotify metadata for "{url}".'.format(url=url))
        metadata = self.spotify.track(url)
        return self.metadata_to_standard_form(metadata)

    def from_query(self, query):
        tracks = self.search(query)["tracks"]["items"]
        if not tracks:
            raise SpotifyMetadataNotFoundError(
                'Spotify returned no tracks for the search query "{}".'.format(
                    query,
                )
            )
        return self.metadata_to_standard_form(tracks[0])

    def search(self, query):
        return self.spotify.search(query)

    def _generate_token(self, client_id, client_secret):
        """ Generate the token. """
        credentials = oauth2.SpotifyClientCredentials(
            client_secret=client_secret,
        )
        token = credentials.get_access_token()
        return token

    def metadata_to_standard_form(self, metadata):
        artist = self.spotify.artist(metadata["artists"][0]["id"])
        album = self.spotify.album(metadata["album"]["id"])

        try:
            metadata[u"genre"] = spotdl.util.titlecase(artist["genres"][0])
        except IndexError:
            metadata[u"genre"] = None
        try:
            metadata[u"copyright"] = album["copyrights"][0]["text"]
        except IndexError:
            metadata[u"copyright"] = None
        try:
            metadata[u"external_ids"][u"isrc"]
        except KeyError:
            metadata[u"external_ids"][u"isrc"] = None

        metadata[u"release_date"] = album["release_date"]
        metadata[u"publisher"] = album["label"]
        metadata[u"total_tracks"] = album["tracks"]["total"]

        # Some sugar
        metadata["year"], *_ = metadata["release_date"].split("-")
        metadata["duration"] = metadata["duration_ms"] / 1000.0
        metadata["provider"] = "spotify"

        # Remove unwanted parameters
        del metadata["duration_ms"]
        del metadata["available_markets"]
        del metadata["album"]["available_markets"]

        return metadata




import lyricwikia

from abc import ABC
from abc import abstractmethod


class LyricBase(ABC):
    """
    Defined lyric providers must inherit from this abstract base
    class and implement their own functionality for the below
    defined methods.
    """

    @abstractmethod
    def from_url(self, url, linesep="\n", timeout=None):
        """
        This method must return the lyrics string for the
        given track.
        """
        pass

    @abstractmethod
    def from_artist_and_track(self, artist, track, linesep="\n", timeout=None):
        """
        This method must return the lyrics string for the
        given track.
        """
        pass

    @abstractmethod
    def from_query(self, query, linesep="\n", timeout=None):
        """
        This method must return the lyrics string for the
        given track.
        """
        pass



from spotdl.metadata.providers import ProviderSpotify
from spotdl.metadata.providers import ProviderYouTube
from spotdl.metadata.providers import YouTubeSearch
from spotdl.metadata.embedders import EmbedderDefault
from spotdl.metadata.exceptions import SpotifyMetadataNotFoundError
import spotdl.metadata

from spotdl.lyrics.providers import LyricWikia
from spotdl.lyrics.providers import Genius
from spotdl.lyrics.exceptions import LyricsNotFoundError

from spotdl.encode.encoders import EncoderFFmpeg

from spotdl.authorize.services import AuthorizeSpotify

from spotdl.track import Track
import spotdl.util
import spotdl.config

from spotdl.command_line.exceptions import NoYouTubeVideoFoundError
from spotdl.command_line.exceptions import NoYouTubeVideoMatchError
from spotdl.metadata_search import MetadataSearch

from spotdl.helpers.spotify import SpotifyHelpers

import sys
import os
import urllib.request

import logging
logger = logging.getLogger(__name__)


class Spotdl:
    def __init__(self, argument_handler):
        self.arguments = argument_handler.run_errands()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        del self

    def match_arguments(self):
        logger.debug("Received arguments:\n{}".format(self.arguments))

        if self.arguments["remove_config"]:
            self.remove_saved_config()
            return
        self.save_default_config()

        AuthorizeSpotify(
            client_id=self.arguments["spotify_client_id"],
            client_secret=self.arguments["spotify_client_secret"]
        )
        spotify_tools = SpotifyHelpers()
        if self.arguments["song"]:
            for track in self.arguments["song"]:
                if track == "-":
                    for line in sys.stdin:
                        self.download_track(
                            line.strip(),
                        )
                else:
                    self.download_track(track)
        elif self.arguments["list"]:
            if self.arguments["write_m3u"]:
                self.write_m3u(
                    self.arguments["list"],
                    self.arguments["write_to"]
                )
            else:
                list_download = {
                    "synchronous": self.download_tracks_from_file,
                    # "threaded"  : self.download_tracks_from_file_threaded,
                }[self.arguments["processor"]]

                list_download(
                    self.arguments["list"],
                )
        elif self.arguments["playlist"]:
            playlist = spotify_tools.fetch_playlist(self.arguments["playlist"])
            spotify_tools.write_playlist_tracks(playlist, self.arguments["write_to"])
        elif self.arguments["album"]:
            album = spotify_tools.fetch_album(self.arguments["album"])
            spotify_tools.write_album_tracks(album, self.arguments["write_to"])
        elif self.arguments["all_albums"]:
            albums = spotify_tools.fetch_albums_from_artist(self.arguments["all_albums"])
            spotify_tools.write_all_albums(albums, self.arguments["write_to"])
        elif self.arguments["username"]:
            playlist_url = spotify_tools.prompt_for_user_playlist(self.arguments["username"])
            playlist = spotify_tools.fetch_playlist(playlist_url)
            spotify_tools.write_playlist_tracks(playlist, self.arguments["write_to"])

    def save_config(self, config_file=spotdl.config.DEFAULT_CONFIG_FILE, config=spotdl.config.DEFAULT_CONFIGURATION):
        config_dir = os.path.dirname(config_file)
        os.makedirs(config_dir, exist_ok=True)
        logger.info('Writing configuration to "{0}":'.format(config_file))
        spotdl.config.dump_config(config_file=config_file, config=spotdl.config.DEFAULT_CONFIGURATION)
        config = spotdl.config.dump_config(config=spotdl.config.DEFAULT_CONFIGURATION["spotify-downloader"])
        for line in config.split("\n"):
            if line.strip():
                logger.info(line.strip())
        logger.info(
            "Please note that command line arguments have higher priority "
            "than their equivalents in the configuration file.\n"
        )

    def save_default_config(self):
        if not os.path.isfile(spotdl.config.DEFAULT_CONFIG_FILE):
            self.save_config()

    def remove_saved_config(self, config_file=spotdl.config.DEFAULT_CONFIG_FILE):
        if os.path.isfile(spotdl.config.DEFAULT_CONFIG_FILE):
            logger.info('Removing "{}".'.format(spotdl.config.DEFAULT_CONFIG_FILE))
            os.remove(spotdl.config.DEFAULT_CONFIG_FILE)
        else:
            logger.info('File does not exist: "{}".'.format(spotdl.config.DEFAULT_CONFIG_FILE))

    def write_m3u(self, track_file, target_file=None):
        with open(track_file, "r") as fin:
            tracks = fin.read().splitlines()

        logger.info(
            "Checking and removing any duplicate tracks in {}.".format(track_file)
        )
        # Remove duplicates and empty elements
        # Also strip whitespaces from elements (if any)
        tracks = spotdl.util.remove_duplicates(
            tracks,
            condition=lambda x: x,
            operation=str.strip
        )

        if target_file is None:
            target_file = "{}.m3u".format(track_file.split(".")[0])

        total_tracks = len(tracks)
        logger.info("Generating {0} from {1} YouTube URLs.".format(target_file, total_tracks))
        write_to_stdout = target_file == "-"
        m3u_headers = "#EXTM3U\n\n"
        if write_to_stdout:
            sys.stdout.write(m3u_headers)
        else:
            with open(target_file, "w") as output_file:
                output_file.write(m3u_headers)

        videos = []
        for n, track in enumerate(tracks, 1):
            try:
                search_metadata = MetadataSearch(
                    track,
                    lyrics=not self.arguments["no_metadata"],
                    yt_search_format=self.arguments["search_format"],
                    yt_manual=self.arguments["manual"]
                )
                video = search_metadata.best_on_youtube_search()
            except (NoYouTubeVideoFoundError, NoYouTubeVideoMatchError) as e:
                logger.error(e.args[0])
            else:
                logger.info(
                    "Matched track {0}/{1} ({2})".format(
                        str(n).zfill(len(str(total_tracks))),
                        total_tracks,
                        video["url"],
                    )
                )
                m3u_key = "#EXTINF:{duration},{title}\n{youtube_url}\n".format(
                    duration=spotdl.util.get_sec(video["duration"]),
                    title=video["title"],
                    youtube_url=video["url"],
                )
                logger.debug(m3u_key.strip())
                if write_to_stdout:
                    sys.stdout.write(m3u_key)
                else:
                    with open(target_file, "a") as output_file:
                        output_file.write(m3u_key)

    def download_track(self, track):
        logger.info('Downloading "{}"'.format(track))
        search_metadata = MetadataSearch(
            track,
            lyrics=not self.arguments["no_metadata"],
            yt_search_format=self.arguments["search_format"],
            yt_manual=self.arguments["manual"]
        )
        try:
            if self.arguments["no_metadata"]:
                metadata = search_metadata.on_youtube()
            else:
                metadata = search_metadata.on_youtube_and_spotify()
        except (NoYouTubeVideoFoundError, NoYouTubeVideoMatchError) as e:
            logger.error(e.args[0])
        else:
            self.download_track_from_metadata(metadata)

    def should_we_overwrite_existing_file(self, overwrite):
        if overwrite == "force":
            logger.info("Forcing overwrite on existing file.")
            to_overwrite = True
        elif overwrite == "prompt":
            to_overwrite = input("Overwrite? (y/N): ").lower() == "y"
        else:
            logger.info("Not overwriting existing file.")
            to_overwrite = False

        return to_overwrite

    def generate_temp_filename(self, filename, for_stdout=False):
        if for_stdout:
            return filename
        return "{filename}.temp".format(filename=filename)

    def output_filename_filter(self, allow_spaces):
        replace_spaces_with_underscores = not allow_spaces
        if replace_spaces_with_underscores:
            return lambda s: s.replace(" ", "_")
        return lambda s: s

    def download_track_from_metadata(self, metadata):
        track = Track(metadata, cache_albumart=(not self.arguments["no_metadata"]))
        stream = metadata["streams"].get(
            quality=self.arguments["quality"],
            preftype=self.arguments["input_ext"],
        )
        if stream is None:
            logger.error('No matching streams found for given input format: "{}".'.format(
                self.arguments["input_ext"]
            ))
            return

        if self.arguments["no_encode"]:
            output_extension = stream["encoding"]
        else:
            output_extension = self.arguments["output_ext"]

        filename = spotdl.metadata.format_string(
            self.arguments["output_file"],
            metadata,
            output_extension=output_extension,
            sanitizer=lambda s: spotdl.util.sanitize(
                s, spaces_to_underscores=self.arguments["no_spaces"]
            )
        )
        download_to_stdout = filename == "-"
        temp_filename = self.generate_temp_filename(filename, for_stdout=download_to_stdout)

        to_skip_download = self.arguments["dry_run"]
        if os.path.isfile(filename):
            logger.info('A file with name "{filename}" already exists.'.format(
                filename=filename
            ))
            to_skip_download = to_skip_download \
                or not self.should_we_overwrite_existing_file(self.arguments["overwrite"])

        if to_skip_download:
            logger.debug("Skip track download.")
            return

        if not self.arguments["no_metadata"]:
            metadata["lyrics"].start()

        os.makedirs(os.path.dirname(filename) or ".", exist_ok=True)

        logger.info('Downloading to "{filename}"'.format(filename=filename))
        if self.arguments["no_encode"]:
            track.download(stream, temp_filename)
        else:
            encoder = EncoderFFmpeg()
            if self.arguments["trim_silence"]:
                encoder.set_trim_silence()
            track.download_while_re_encoding(
                stream,
                temp_filename,
                target_encoding=output_extension,
                encoder=encoder,
            )

        if not self.arguments["no_metadata"]:
            track.metadata["lyrics"] = track.metadata["lyrics"].join()
            self.apply_metadata(track, temp_filename, output_extension)

        if not download_to_stdout:
            logger.debug("Renaming {temp_filename} to {filename}.".format(
                temp_filename=temp_filename, filename=filename
            ))
            os.rename(temp_filename, filename)

        return filename

    def apply_metadata(self, track, filename, encoding):
        logger.info("Applying metadata")
        try:
            track.apply_metadata(filename, encoding=encoding)
        except TypeError:
            logger.warning("Cannot apply metadata on provided output format.")

    def strip_and_filter_duplicates(self, tracks):
        filtered_tracks = spotdl.util.remove_duplicates(
            tracks,
            condition=lambda x: x,
            operation=str.strip
        )
        return filtered_tracks

    def filter_against_skip_file(self, items, skip_file):
        skip_items = spotdl.util.readlines_from_nonbinary_file(skip_file)
        filtered_skip_items = self.strip_and_filter_duplicates(skip_items)
        filtered_items = [item for item in items if not item in filtered_skip_items]
        return filtered_items

    def download_tracks_from_file(self, path):
        logger.info(
            'Checking and removing any duplicate tracks in "{}".'.format(path)
        )
        tracks = spotdl.util.readlines_from_nonbinary_file(path)
        tracks = self.strip_and_filter_duplicates(tracks)

        if self.arguments["skip_file"]:
            len_tracks_before = len(tracks)
            tracks = self.filter_against_skip_file(tracks, self.arguments["skip_file"])
            logger.info("Skipping {} tracks due to matches in skip file.".format(
                len_tracks_before - len(tracks))
            )
        # Overwrite file
        spotdl.util.writelines_to_nonbinary_file(path, tracks)

        logger.info(
            "Downloading {n} tracks.\n".format(n=len(tracks))
        )

        for position, track in enumerate(tracks, 1):
            search_metadata = MetadataSearch(
                track,
                lyrics=True,
                yt_search_format=self.arguments["search_format"],
                yt_manual=self.arguments["manual"]
            )
            log_track_query = '{position}. Downloading "{track}"'.format(
                position=position,
                track=track
            )
            logger.info(log_track_query)
            try:
                metadata = search_metadata.on_youtube_and_spotify()
                self.download_track_from_metadata(metadata)
            except (urllib.request.URLError, TypeError, IOError) as e:
                logger.exception(e.args[0])
                logger.warning(
                    "Failed to download current track due to possible network issue. "
                    "Will retry after other songs."
                )
                tracks.append(track)
            except (NoYouTubeVideoFoundError, NoYouTubeVideoMatchError) as e:
                logger.error("{err}".format(err=e.args[0]))
            except KeyboardInterrupt:
                # The current track hasn't been downloaded completely.
                # Make sure we continue from here the next the program runs.
                tracks.insert(0, track)
                raise
            else:
                if self.arguments["write_successful_file"]:
                    with open(self.arguments["write_successful_file"], "a") as fout:
                        fout.write("{}\n".format(track))
            finally:
                spotdl.util.writelines_to_nonbinary_file(path, tracks[position:])
                print("", file=sys.stderr)

    """
    def download_tracks_from_file_threaded(self, path):
        # FIXME: Can we make this function cleaner?

        logger.info(
            "Checking and removing any duplicate tracks in {}.\n".format(path)
        )
        with open(path, "r") as fin:
            # Read tracks into a list and remove any duplicates
            tracks = fin.read().splitlines()

        # Remove duplicates and empty elements
        # Also strip whitespaces from elements (if any)
        spotdl.util.remove_duplicates(
            tracks,
            condition=lambda x: x,
            operation=str.strip
        )

        # Overwrite file
        with open(path, "w") as fout:
            fout.writelines(tracks)

        tracks_count = len(tracks)
        current_iteration = 1

        next_track = tracks.pop(0)
        metadata = {
            "current_track": None,
            "next_track": spotdl.util.ThreadWithReturnValue(
                target=search_metadata,
                args=(next_track, self.arguments["search_format"])
            )
        }
        metadata["next_track"].start()
        while tracks_count > 0:
            metadata["current_track"] = metadata["next_track"].join()
            metadata["next_track"] = None
            try:
                print(tracks_count, file=sys.stderr)
                print(tracks, file=sys.stderr)
                if tracks_count > 1:
                    current_track = next_track
                    next_track = tracks.pop(0)
                    metadata["next_track"] = spotdl.util.ThreadWithReturnValue(
                        target=search_metadata,
                        args=(next_track, self.arguments["search_format"])
                    )
                    metadata["next_track"].start()

                log_track_query = str(current_iteration) + ". {artist} - {track-name}"
                logger.info(log_track_query)
                if metadata["current_track"] is None:
                    logger.warning("Something went wrong. Will retry after downloading remaining tracks.")
                    pass
                print(metadata["current_track"]["name"], file=sys.stderr)
                # self.download_track_from_metadata(metadata["current_track"])
            except (urllib.request.URLError, TypeError, IOError) as e:
                print("", file=sys.stderr)
                logger.exception(e.args[0])
                logger.warning("Failed. Will retry after other songs\n")
                tracks.append(current_track)
            else:
                tracks_count -= 1
                if self.arguments["write_sucessful_file"]:
                    with open(self.arguments["write_sucessful_file"], "a") as fout:
                        fout.write(current_track)
            finally:
                current_iteration += 1
                with open(path, "w") as fout:
                    fout.writelines(tracks)
    """




class EncoderNotFoundError(Exception):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)


class FFmpegNotFoundError(EncoderNotFoundError):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)




import logging
import coloredlogs

import sys

from spotdl.command_line.core import Spotdl
from spotdl.command_line.arguments import get_arguments
from spotdl.command_line.exceptions import ArgumentError

# hardcode loglevel for dependencies so that they do not spew generic
# log messages along with spotdl.
for module in ("chardet", "urllib3", "spotipy", "pytube"):
    logging.getLogger(module).setLevel(logging.CRITICAL)

coloredlogs.DEFAULT_FIELD_STYLES = {
    "levelname": {"bold": True, "color": "yellow"},
    "name": {"color": "blue"},
    "lineno": {"color": "magenta"},
}


def set_logger(level):
    if level == logging.DEBUG:
        fmt = "%(levelname)s:%(name)s:%(lineno)d:\n%(message)s\n"
    else:
        fmt = "%(levelname)s: %(message)s"
    logging.basicConfig(format=fmt, level=level)
    logger = logging.getLogger(name=__name__)
    coloredlogs.install(level=level, fmt=fmt, logger=logger)
    return logger


def main():
    try:
        argument_handler = get_arguments()
    except ArgumentError as e:
        logger = set_logger(logging.INFO)
        logger.info(e.args[0])
        sys.exit(5)

    logging_level = argument_handler.get_logging_level()
    logger = set_logger(logging_level)
    try:
        spotdl = Spotdl(argument_handler)
    except ArgumentError as e:
        argument_handler.parser.error(e.args[0])
    try:
        spotdl.match_arguments()
    except KeyboardInterrupt as e:
        print("", file=sys.stderr)
        logger.exception(e)
        sys.exit(2)


if __name__ == "__main__":
    main()




import pytube
from bs4 import BeautifulSoup

import urllib.request
import threading
from collections.abc import Sequence

from spotdl.metadata import StreamsBase
from spotdl.metadata import ProviderBase
from spotdl.metadata.exceptions import YouTubeMetadataNotFoundError

import spotdl.util

import logging
logger = logging.getLogger(__name__)

BASE_SEARCH_URL = "https://www.youtube.com/results?sp=EgIQAQ%253D%253D&q={}"
HEADERS = [('Range', 'bytes=0-'),]


class YouTubeVideos(Sequence):
    def __init__(self, videos):
        self.videos = videos
        super().__init__()

    def __repr__(self):
        return "YouTubeVideos({})".format(self.videos)

    def __len__(self):
        return len(self.videos)

    def __getitem__(self, index):
        return self.videos[index]

    def __eq__(self, instance):
        return self.videos == instance.videos

    def bestmatch(self):
        video = self.videos[0]
        logger.debug("Matched with: {title} ({url}) [{duration}]".format(
            title=video["title"],
            url=video["url"],
            duration=video["duration"]
        ))
        return video


class YouTubeSearch:
    def __init__(self):
        self.base_search_url = BASE_SEARCH_URL

    def generate_search_url(self, query):
        quoted_query = urllib.request.quote(query)
        return self.base_search_url.format(quoted_query)

    def _fetch_response_html(self, url):
        response = urllib.request.urlopen(url)
        soup = BeautifulSoup(response.read(), "html.parser")
        return soup

    def _extract_video_details_from_result(self, html):
        video_time = html.find("span", class_="video-time").get_text()
        inner_html = html.find("div", class_="yt-lockup-content")
        video_id = inner_html.find("a")["href"][-11:]
        video_title = inner_html.find("a")["title"]
        video_details = {
            "url": "https://www.youtube.com/watch?v=" + video_id,
            "title": video_title,
            "duration": video_time,
        }
        return video_details

    def _fetch_search_results(self, html, limit=10):
        result_source = html.find_all(
            "div", {"class": "yt-lockup-dismissable yt-uix-tile"}
        )
        videos = []

        for result in result_source:
            if not self._is_video(result):
                continue

            video = self._extract_video_details_from_result(result)
            videos.append(video)

            if len(videos) >= limit:
                break

        return videos

    def _is_video(self, result):
        # ensure result is not a channel
        not_video = (
            result.find("channel") is not None
            or "yt-lockup-channel" in result.parent.attrs["class"]
            or "yt-lockup-channel" in result.attrs["class"]
        )

        # ensure result is not a mix/playlist
        not_video = not_video or "yt-lockup-playlist" in result.parent.attrs["class"]

        # ensure video result is not an advertisement
        not_video = not_video or result.find("googleads") is not None
        # ensure video result is not a live stream
        not_video = not_video or result.find("span", class_="video-time") is None

        video = not not_video
        return video

    def _is_server_side_invalid_response(self, videos, html):
        if videos:
            return False
        search_message = html.find("div", {"class":"search-message"})
        return search_message is None

    def search(self, query, limit=10, retries=5):
        """ Search and scrape YouTube to return a list of matching videos. """
        search_url = self.generate_search_url(query)
        logger.debug('Fetching YouTube results for "{}" at "{}".'.format(query, search_url))
        html = self._fetch_response_html(search_url)
        videos = self._fetch_search_results(html, limit=limit)
        to_retry = retries > 0 and self._is_server_side_invalid_response(videos, html)
        if to_retry:
            logger.debug(
                "Retrying since YouTube returned invalid response for search "
                "results. Retries left: {retries}.".format(retries=retries)
            )
            return self.search(query, limit=limit, retries=retries-1)
        return YouTubeVideos(videos)


class YouTubeStreams(StreamsBase):
    def __init__(self, streams):
        self.network_headers = HEADERS

        audiostreams = streams.filter(only_audio=True).order_by("abr").desc()

        thread_pool = []
        self.all = []

        for stream in audiostreams:
            encoding = "m4a" if "mp4a" in stream.audio_codec else stream.audio_codec
            standard_stream = {
                # Store only the integer part for bitrate. For example
                # the given bitrate would be "192kbps", we store only
                # the integer part (192) here and drop the rest.
                "bitrate": int(stream.abr[:-4]),
                "connection": None,
                "download_url": stream.url,
                "encoding": encoding,
                "filesize": None,
            }
            establish_connection = threading.Thread(
               target=self._store_connection,
               args=(standard_stream,),
            )
            thread_pool.append(establish_connection)
            establish_connection.start()
            self.all.append(standard_stream)

        for thread in thread_pool:
            thread.join()

    def _store_connection(self, stream):
        response = self._make_request(stream["download_url"])
        stream["connection"] = response
        stream["filesize"] = int(response.headers["Content-Length"])

    def _make_request(self, url):
        request = urllib.request.Request(url)
        for header in self.network_headers:
            request.add_header(*header)
        return urllib.request.urlopen(request)

    def get(self, quality="best", preftype="automatic"):
        if quality == "best":
            return self.getbest(preftype=preftype)
        elif quality == "worst":
            return self.getworst(preftype=preftype)
        else:
            return None

    def getbest(self, preftype="automatic"):
        selected_stream = None
        if preftype == "automatic":
            selected_stream = self.all[0]
        else:
            for stream in self.all:
                if stream["encoding"] == preftype:
                    selected_stream = stream
                    break
        logger.debug('Selected best quality stream for "{preftype}" format:\n{stream}'.format(
            preftype=preftype,
            stream=selected_stream,
        ))
        return selected_stream

    def getworst(self, preftype="automatic"):
        selected_stream = None
        if preftype == "automatic":
            selected_stream = self.all[-1]
        else:
            for stream in self.all[::-1]:
                if stream["encoding"] == preftype:
                    selected_stream = stream
                    break
        logger.debug('Selected worst quality stream for "{preftype}" format:\n{stream}'.format(
            preftype=preftype,
            stream=selected_stream,
        ))
        return selected_stream


class ProviderYouTube(ProviderBase):
    def from_query(self, query):
        watch_urls = self.search(query)
        if not watch_urls:
            raise YouTubeMetadataNotFoundError(
                'YouTube returned nothing for the given search '
                'query ("{}")'.format(query)
            )
        return self.from_url(watch_urls[0])

    def from_url(self, url, retries=5):
        logger.debug('Fetching YouTube metadata for "{url}".'.format(url=url))
        try:
            content = pytube.YouTube(url)
        except KeyError:
            # Sometimes YouTube can return unexpected response, in such a case
            # retry a few times before finally failing.
            if retries > 0:
                retries -= 1
                logger.debug(
                    "YouTube returned an unexpected response for "
                    "`pytube.YouTube({url})`. Retries left: {retries}".format(
                        url=url, retries=retries
                    )
                )
                return self.from_url(url, retries=retries)
            else:
                raise
        else:
            return self.from_pytube_object(content)

    def from_pytube_object(self, content):
        return self.metadata_to_standard_form(content)

    def search(self, query):
        return YouTubeSearch().search(query)

    def _fetch_publish_date(self, content):
        # XXX: This needs to be supported in PyTube itself
        # See https://github.com/nficano/pytube/issues/595
        position = content.watch_html.find("publishDate")
        publish_date = content.watch_html[position+16:position+25]
        return publish_date

    def metadata_to_standard_form(self, content):
        """ Fetch a song's metadata from YouTube. """
        publish_date = self._fetch_publish_date(content)
        metadata = {
            "name": content.title,
            "artists": [{"name": content.author}],
            "duration": content.length,
            "external_urls": {"youtube": content.watch_url},
            "album": {
                "images": [{"url": content.thumbnail_url}],
                "artists": [{"name": None}],
                "name": None,
            },
            "year": publish_date.split("-")[0],
            "release_date": publish_date,
            "type": "track",
            "disc_number": 1,
            "track_number": 1,
            "total_tracks": 1,
            "publisher": None,
            "external_ids": {"isrc": None},
            "lyrics": None,
            "copyright": None,
            "genre": None,
            "streams": YouTubeStreams(content.streams),
            "provider": "youtube",
        }
        return metadata



import appdirs

import argparse
import mimetypes
import os
import sys
import shutil

from spotdl.command_line.exceptions import ArgumentError
import spotdl.util
import spotdl.config

from collections.abc import Sequence
import logging
logger = logging.getLogger(__name__)


_LOG_LEVELS = {
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "DEBUG": logging.DEBUG,
}

if os.path.isfile(spotdl.config.DEFAULT_CONFIG_FILE):
    saved_config = spotdl.config.read_config(spotdl.config.DEFAULT_CONFIG_FILE)
else:
    saved_config = {"spotify-downloader": {}}

_CONFIG_BASE = spotdl.util.merge_copy(
    spotdl.config.DEFAULT_CONFIGURATION,
    saved_config,
)


def get_arguments(config_base=_CONFIG_BASE):
    parser = argparse.ArgumentParser(
        description="Download and convert tracks from Spotify, Youtube, etc.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    defaults = config_base["spotify-downloader"]

    to_remove_config = "--remove-config" in sys.argv[1:]
    if not to_remove_config and "download-only-metadata" in defaults:
        raise ArgumentError(
            "The default configuration file currently set is not suitable for spotdl>=2.0.0.\n"
            "You need to remove your previous `config.yml` due to breaking changes\n"
            "introduced in v2.0.0, new options being added, and old ones being removed\n"
            "You may want to first backup your old configuration for reference. You can\n"
            "then remove your current configuration by running:\n"
            "```\n"
            "$ spotdl --remove-config\n"
            "```\n"
            "spotdl will automatically generate a new configuration file on the next run.\n"
            "You can then replace the appropriate fields in the newly generated\n"
            "configuration file by referring to your old configuration file.\n\n"
            "For the list of OTHER BREAKING CHANGES and release notes check out:\n"
            "https://github.com/ritiek/spotify-downloader/releases/tag/v2.0.0"
        )

    possible_special_tags = (
        "{track-name}",
        "{artist}",
        "{album}",
        "{album-artist}",
        "{genre}",
        "{disc-number}",
        "{duration}",
        "{year}",
        "{original-date}",
        "{track-number}",
        "{total-tracks}",
        "{isrc}",
        "{track-id}",
        "{output-ext}",
    )

    # `--remove-config` does not require the any of the group arguments to be passed.
    group = parser.add_mutually_exclusive_group(required=not to_remove_config)

    group.add_argument(
        "-s",
        "--song",
        nargs="+",
        help="download track(s) by spotify link or name"
    )
    group.add_argument(
        "-l",
        "--list",
        help="download tracks from a file (WARNING: this file will be modified!)"
    )
    group.add_argument(
        "-p",
        "--playlist",
        help="load tracks from playlist URL into <playlist_name>.txt or "
             "if `--write-to=<path/to/file.txt>` has been passed",
    )
    group.add_argument(
        "-a",
        "--album",
        help="load tracks from album URL into <album_name>.txt or if "
             "`--write-to=<path/to/file.txt>` has been passed"
    )
    group.add_argument(
        "-aa",
        "--all-albums",
        help="load all tracks from artist URL into <artist_name>.txt "
             "or if `--write-to=<path/to/file.txt>` has been passed"
    )
    group.add_argument(
        "-u",
        "--username",
        help="load tracks from user's playlist into <playlist_name>.txt "
             "or if `--write-to=<path/to/file.txt>` has been passed"
    )

    parser.add_argument(
        "--write-m3u",
        help="generate an .m3u playlist file with youtube links given "
             "a text file containing tracks",
        action="store_true",
    )
    parser.add_argument(
        "-m",
        "--manual",
        default=defaults["manual"],
        help="choose the track to download manually from a list of matching tracks",
        action="store_true",
    )
    parser.add_argument(
        "-nm",
        "--no-metadata",
        default=defaults["no_metadata"],
        help="do not embed metadata in tracks",
        action="store_true",
    )
    parser.add_argument(
        "-ne",
        "--no-encode",
        default=defaults["no_encode"],
        action="store_true",
        help="do not encode media using FFmpeg",
    )
    parser.add_argument(
        "--overwrite",
        default=defaults["overwrite"],
        choices={"prompt", "force", "skip"},
        help="change the overwrite policy",
    )
    parser.add_argument(
        "-q",
        "--quality",
        default=defaults["quality"],
        choices={"best", "worst"},
        help="preferred audio quality",
    )
    parser.add_argument(
        "-i",
        "--input-ext",
        default=defaults["input_ext"],
        choices={"automatic", "m4a", "opus"},
        help="preferred input format",
    )
    parser.add_argument(
        "-o",
        "--output-ext",
        default=defaults["output_ext"],
        choices={"mp3", "m4a", "flac"},
        help="preferred output format",
    )
    parser.add_argument(
        "--write-to",
        default=defaults["write_to"],
        help="write tracks from Spotify playlist, album, etc. to this file",
    )
    parser.add_argument(
        "-f",
        "--output-file",
        default=defaults["output_file"],
        help="path where to write the downloaded track to, special tags "
        "are to be surrounded by curly braces. Possible tags: {}".format(
            possible_special_tags
        )
    )
    parser.add_argument(
        "--trim-silence",
        default=defaults["trim_silence"],
        help="remove silence from the start of the audio",
        action="store_true",
    )
    parser.add_argument(
        "-sf",
        "--search-format",
        default=defaults["search_format"],
        help="search format to search for on YouTube, special tags "
        "are to be surrounded by curly braces. Possible tags: {}".format(
            possible_special_tags
        )
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        default=defaults["dry_run"],
        help="show only track title and YouTube URL, and then skip "
        "to the next track (if any)",
        action="store_true",
    )
    parser.add_argument(
        "--processor",
        default="synchronous",
        choices={"synchronous", "threaded"},
        # help='list downloading strategy: - "synchronous" downloads '
        # 'tracks one-by-one. - "threaded" (highly experimental at the '
        # 'moment! expect it to slash & burn) pre-fetches the next '
        # 'track\'s metadata for more efficient downloading'
        # XXX: Still very experimental to be exposed
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-ns",
        "--no-spaces",
        default=defaults["no_spaces"],
        help="replace spaces in metadata values with underscores when "
        "generating filenames",
        action="store_true",
    )
    parser.add_argument(
        "-sk",
        "--skip-file",
        default=defaults["skip_file"],
        help="path to file containing tracks to skip",
    )
    parser.add_argument(
        "-w",
        "--write-successful-file",
        default=defaults["write_successful_file"],
        help="path to file to write successful tracks to",
    )
    parser.add_argument(
        "--spotify-client-id",
        default=defaults["spotify_client_id"],
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--spotify-client-secret",
        default=defaults["spotify_client_secret"],
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-ll",
        "--log-level",
        default=defaults["log_level"],
        choices=_LOG_LEVELS.keys(),
        type=str.upper,
        help="set log verbosity",
    )
    parser.add_argument(
        "-c",
        "--config",
        default=spotdl.config.DEFAULT_CONFIG_FILE,
        help="path to custom config.yml file"
    )
    parser.add_argument(
        "--remove-config",
        default=False,
        action="store_true",
        help="remove previously saved config"
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {}".format(spotdl.__version__),
    )

    return ArgumentHandler(parser=parser)


class ArgumentHandler:
    def __init__(self, args=None, parser=argparse.ArgumentParser(""), config_base=_CONFIG_BASE):
        args_were_passed = args is not None
        if not args_were_passed:
            args = parser.parse_args().__dict__

        config_file = args.get("config")
        configured_args = args.copy()
        if config_file and os.path.isfile(config_file):
            config = spotdl.config.read_config(config_file)
            parser.set_defaults(**config["spotify-downloader"])
            configured_args = parser.parse_args().__dict__

        if args_were_passed:
            parser.set_defaults(**args)
            configured_args = parser.parse_args().__dict__

        defaults = config_base["spotify-downloader"]
        args = spotdl.util.merge_copy(defaults, args)

        self.parser = parser
        self.args = args
        self.configured_args = configured_args

    def get_configured_args(self):
        return self.configured_args

    def get_logging_level(self):
        return _LOG_LEVELS[self.configured_args["log_level"]]

    def run_errands(self):
        args = self.get_configured_args()

        if (args.get("list")
            and not mimetypes.MimeTypes().guess_type(args["list"])[0] == "text/plain"
        ):
            raise ArgumentError(
                "{0} is not of a valid argument to --list, argument must be plain text file.".format(
                    args["list"]
                )
            )

        if args.get("write_m3u") and not args.get("list"):
            raise ArgumentError("--write-m3u can only be used with --list.")

        if args["write_to"] and not (
            args.get("playlist") or args.get("album") or args.get("all_albums") or args.get("username") or args.get("write_m3u")
        ):
            raise ArgumentError(
                "--write-to can only be used with --playlist, --album, --all-albums, --username, or --write-m3u."
            )

        ffmpeg_exists = shutil.which("ffmpeg")
        if not ffmpeg_exists:
            logger.warn("FFmpeg was not found in PATH. Will not re-encode media to specified output format.")
            args["no_encode"] = True

        if args["no_encode"] and args["trim_silence"]:
            logger.warn("--trim-silence can only be used when an encoder is set.")

        if args["output_file"] == "-" and args["no_metadata"] is False:
            logger.warn(
                "Cannot write metadata when target is STDOUT. Pass "
                "--no-metadata explicitly to hide this warning."
            )
            args["no_metadata"] = True
        elif os.path.isdir(args["output_file"]):
            adjusted_output_file = os.path.join(
                args["output_file"],
                self.parser.get_default("output_file")
            )
            logger.warn(
                "Given output file is a directory. Will download tracks "
                "in this directory with their filename as per the default "
                "file format. Pass --output-file=\"{}\" to hide this "
                "warning.".format(
                    adjusted_output_file
                )
            )
            args["output_file"] = adjusted_output_file

        return args




from spotdl.authorize import AuthorizeBase
from spotdl.authorize.exceptions import SpotifyAuthorizationError

import spotipy
import spotipy.oauth2 as oauth2

import logging
logger = logging.getLogger(__name__)

# This masterclient is used to keep the last logged-in client
# object in memory for for persistence. If credentials aren't
# provided when creating further objects, the last authenticated
# client object with correct credentials is returned when
# `AuthorizeSpotify().authorize()` is called.
masterclient = None

class AuthorizeSpotify(spotipy.Spotify):
    def __init__(self, client_id=None, client_secret=None):
        global masterclient

        credentials_provided = client_id is not None \
                           and client_secret is not None
        valid_input = credentials_provided or masterclient is not None

        if not valid_input:
            raise SpotifyAuthorizationError(
                "You must pass in client_id and client_secret to this method "
                "when authenticating for the first time."
            )

        if masterclient:
            logger.debug("Reading cached master Spotify credentials.")
            # Use cached client instead of authorizing again
            # and thus wasting time.
            self.__dict__.update(masterclient.__dict__)
        else:
            logger.debug("Setting master Spotify credentials.")
            credential_manager = oauth2.SpotifyClientCredentials(
                client_id=client_id,
                client_secret=client_secret
            )
            super().__init__(client_credentials_manager=credential_manager)
            # Cache current client
            masterclient = self




import shutil
import os

from abc import ABC
from abc import abstractmethod

from spotdl.encode.exceptions import EncoderNotFoundError

"""
  NOTE ON ENCODERS
  ================

* FFmeg encoders sorted in descending order based
  on the quality of audio produced:
  libopus > libvorbis >= libfdk_aac > aac > libmp3lame

* libfdk_aac encoder, due to copyrights needs to be compiled
  by end user on MacOS brew install ffmpeg --with-fdk-aac
  will do just that. Other OS? See:
  https://trac.ffmpeg.org/wiki/Encode/AAC

"""

_TARGET_FORMATS_FROM_ENCODING = {
    "m4a": "mp4",
    "mp3": "mp3",
    "opus": "opus",
    "flac": "flac"
}


class EncoderBase(ABC):
    """
    Defined encoders must inherit from this abstract base class
    and implement their own functionality for the below defined
    methods.
    """

    @abstractmethod
    def __init__(self, encoder_path, must_exist, loglevel, additional_arguments=[]):
        """
        This method must make sure whether specified encoder
        is available under PATH.
        """
        if must_exist and shutil.which(encoder_path) is None:
            raise EncoderNotFoundError(
                "{} executable does not exist or was not found in PATH.".format(
                    encoder_path
                )
            )
        self.encoder_path = encoder_path
        self._loglevel = loglevel
        self._additional_arguments = additional_arguments
        self._target_formats_from_encoding = _TARGET_FORMATS_FROM_ENCODING

    def set_argument(self, argument):
        """
        This method must be used to set any custom functionality
        for the encoder by passing arguments to it.
        """
        self._additional_arguments += argument.split()

    def get_encoding(self, path):
        """
        This method must determine the encoding for a local
        audio file. Such as "mp3", "wav", "m4a", etc.
        """
        _, extension = os.path.splitext(path)
        # Ignore the initial dot from file extension
        return extension[1:]

    @abstractmethod
    def set_debuglog(self):
        """
        This method must enable verbose logging in the defined
        encoder.
        """
        pass

    @abstractmethod
    def _generate_encode_command(self, input_path, target_path):
        """
        This method must the complete command for that would be
        used to invoke the encoder and perform the encoding.
        """
        pass

    @abstractmethod
    def _generate_encoding_arguments(self, input_encoding, target_encoding):
        """
        This method must return the core arguments for the defined
        encoder such as defining the sample rate, audio bitrate,
        etc.
        """
        pass

    @abstractmethod
    def re_encode(self, input_path, target_path):
        """
        This method must invoke the encoder to encode a given input
        file to a specified output file.
        """
        pass

    def target_format_from_encoding(self, encoding):
        """
        This method generates the target stream format from given
        input encoding.
        """
        target_format = self._target_formats_from_encoding[encoding]
        return target_format

    def re_encode_from_stdin(self, input_encoding, target_path):
        """
        This method must invoke the encoder to encode stdin to a
        specified output file.
        """
        raise NotImplementedError




from spotdl.authorize.services import AuthorizeSpotify
import spotdl.util

import sys
import spotipy

import logging
logger = logging.getLogger(__name__)

try:
    from slugify import SLUG_OK, slugify
except ImportError:
    logger.error("Oops! `unicode-slugify` was not found.")
    logger.info("Please remove any other slugify library and install `unicode-slugify`.")
    raise


class SpotifyHelpers:
    def __init__(self, spotify=None):
        if spotify is None:
            spotify = AuthorizeSpotify()
        self.spotify = spotify

    def prompt_for_user_playlist(self, username):
        """ Write user playlists to target_file """
        playlists = self.fetch_user_playlist_urls(username)
        for i, playlist in enumerate(playlists, 1):
            playlist_details = "{0}. {1:<30}  ({2} tracks)".format(
                i, playlist["name"], playlist["tracks"]["total"]
            )
            print(playlist_details, file=sys.stderr)
        print("", file=sys.stderr)
        playlist = spotdl.util.prompt_user_for_selection(playlists)
        return playlist["external_urls"]["spotify"]

    def fetch_user_playlist_urls(self, username):
        """ Fetch user playlists when using the -u option. """
        logger.debug('Fetching playlists for "{username}".'.format(username=username))
        playlists = self.spotify.user_playlists(username)
        collected_playlists = []
        check = 1

        while True:
            for playlist in playlists["items"]:
                # in rare cases, playlists may not be found, so playlists['next']
                # is None. Skip these. Also see Issue #91.
                if playlist["name"] is not None:
                    collected_playlists.append(playlist)
                    check += 1
            if playlists["next"]:
                playlists = self.spotify.next(playlists)
            else:
                break

        return collected_playlists

    def fetch_playlist(self, playlist_url):
        logger.debug('Fetching playlist "{playlist}".'.format(playlist=playlist_url))
        try:
            results = self.spotify.playlist(playlist_url, fields="tracks,next,name")
        except spotipy.client.SpotifyException:
            logger.exception(
                "Unable to find playlist. Make sure the playlist is set "
                "to publicly visible and then try again."
            )

        return results

    def write_playlist_tracks(self, playlist, target_file=None):
        tracks = playlist["tracks"]
        if not target_file:
            target_file = u"{0}.txt".format(slugify(playlist["name"], ok="-_()[]{}"))
        return self.write_tracks(tracks, target_file)

    def fetch_album(self, album_uri):
        logger.debug('Fetching album "{album}".'.format(album=album_uri))
        album = self.spotify.album(album_uri)
        return album

    def write_album_tracks(self, album, target_file=None):
        tracks = self.spotify.album_tracks(album["id"])
        if not target_file:
            target_file = u"{0}.txt".format(slugify(album["name"], ok="-_()[]{}"))
        return self.write_tracks(tracks, target_file)

    def fetch_albums_from_artist(self, artist_uri, album_type=None):
        """
        This function returns all the albums from a give artist_uri using the US
        market
        :param artist_uri - spotify artist uri
        :param album_type - the type of album to fetch (ex: single) the default is
                            all albums
        :param return - the album from the artist
        """

        logger.debug('Fetching all albums for "{artist}".'.format(artist=artist_uri))
        # fetching artist's albums limitting the results to the US to avoid duplicate
        # albums from multiple markets
        results = self.spotify.artist_albums(artist_uri, album_type=album_type, country="US")

        albums = results["items"]

        # indexing all pages of results
        while results["next"]:
            results = self.spotify.next(results)
            albums.extend(results["items"])

        return albums

    def write_all_albums(self, albums, target_file=None):
        """
        This function gets all albums from an artist and writes it to a file in the
        current working directory called [ARTIST].txt, where [ARTIST] is the artist
        of the album
        :param artist_uri - spotify artist uri
        :param target_file - file to write albums to
        """

        # if no file if given, the default save file is in the current working
        # directory with the name of the artist
        if target_file is None:
            target_file = albums[0]["artists"][0]["name"] + ".txt"

        for album in albums:
            logger.info('Fetching album "{album}".'.format(album=album["name"]))
            self.write_album_tracks(album, target_file=target_file)

    def write_tracks(self, tracks, target_file):
        def writer(tracks, file_io):
            track_urls = []
            while True:
                for item in tracks["items"]:
                    if "track" in item:
                        track = item["track"]
                    else:
                        track = item
                    # Spotify sometimes returns additional empty "tracks" in the API
                    # response. We need to discard such "tracks".
                    # See https://github.com/spotify/web-api/issues/1562
                    if track is None:
                        continue
                    try:
                        track_url = track["external_urls"]["spotify"]
                        file_io.write(track_url + "\n")
                        track_urls.append(track_url)
                    except KeyError:
                        # FIXME: Write "{artist} - {name}" instead of Spotify URI for
                        #        "local only" tracks.
                        logger.warning(
                            'Skipping track "{0}" by "{1}" (local only?)'.format(
                                track["name"], track["artists"][0]["name"]
                            )
                        )
                # 1 page = 50 results
                # check if there are more pages
                if tracks["next"]:
                    tracks = self.spotify.next(tracks)
                else:
                    break
            return track_urls

        logger.info(u"Writing {0} tracks to {1}.".format(tracks["total"], target_file))
        write_to_stdout = target_file == "-"
        if write_to_stdout:
            file_out = sys.stdout
            track_urls = writer(tracks, file_out)
        else:
            with open(target_file, "a") as file_out:
                track_urls = writer(tracks, file_out)
        return track_urls




import os

from abc import ABC
from abc import abstractmethod

import urllib.request

class EmbedderBase(ABC):
    """
    The subclass must define the supported media file encoding
    formats here using a static variable - such as:

    >>> supported_formats = ("mp3", "m4a", "flac")
    """
    supported_formats = ()

    @abstractmethod
    def __init__(self):
        """
        For every supported format, there must be a corresponding
        method that applies metadata on this format.

        Such as if mp3 is supported, there must exist a method named
        `as_mp3` on this class that applies metadata on mp3 files.
        """
        # self.targets = { fmt: eval(str("self.as_" + fmt))
        #                  for fmt in self.supported_formats }
        #
        # TODO: The above code seems to fail for some reason
        #       I do not know.
        self.targets = {}
        for fmt in self.supported_formats:
            # FIXME: Calling `eval` is dangerous here!
            self.targets[fmt] = eval("self.as_" + fmt)

    def get_encoding(self, path):
        """
        This method must determine the encoding for a local
        audio file. Such as "mp3", "wav", "m4a", etc.
        """
        _, extension = os.path.splitext(path)
        # Ignore the initial dot from file extension
        return extension[1:]

    def apply_metadata(self, path, metadata, cached_albumart=None, encoding=None):
        """
        This method must automatically detect the media encoding
        format from file path and embed the corresponding metadata
        on the given file by calling an appropriate submethod.
        """
        if cached_albumart is None:
            cached_albumart = urllib.request.urlopen(
                metadata["album"]["images"][0]["url"],
            ).read()
        if encoding is None:
            encoding = self.get_encoding(path)
        if encoding not in self.supported_formats:
            raise TypeError(
                'The input format ("{}") is not supported.'.format(
                encoding,
            ))
        embed_on_given_format = self.targets[encoding]
        embed_on_given_format(path, metadata, cached_albumart=cached_albumart)

    def as_mp3(self, path, metadata, cached_albumart=None):
        """
        Method for mp3 support. This method might be defined in
        a subclass.

        Other methods for additional supported formats must also
        be declared here.
        """
        raise NotImplementedError

    def as_m4a(self, path, metadata, cached_albumart=None):
        """
        Method for m4a support. This method might be defined in
        a subclass.

        Other methods for additional supported formats must also
        be declared here.
        """
        raise NotImplementedError

    def as_flac(self, path, metadata, cached_albumart=None):
        """
        Method for flac support. This method might be defined in
        a subclass.

        Other methods for additional supported formats must also
        be declared here.
        """
        raise NotImplementedError




import lyricwikia

from spotdl.lyrics.lyric_base import LyricBase
from spotdl.lyrics.exceptions import LyricsNotFoundError


class LyricWikia(LyricBase):
    def from_query(self, query, linesep="\n", timeout=None):
        raise NotImplementedError

    def from_artist_and_track(self, artist, track, linesep="\n", timeout=None):
        """
        Returns the lyric string for the given artist and track.
        """
        try:
            lyrics = lyricwikia.get_lyrics(artist, track, linesep, timeout)
        except lyricwikia.LyricsNotFound as e:
            raise LyricsNotFoundError(e.args[0])

        return lyrics

    def from_url(self, url, linesep="\n", timeout=None):
        raise NotImplementedError




from setuptools import setup
import os

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

# __version__ comes into namespace from here
with open(os.path.join("spotdl", "version.py")) as version_file:
    exec(version_file.read())

setup(
    # 'spotify-downloader' was already taken :/
    name="spotdl",
    # Tests are included automatically:
    # https://docs.python.org/3.6/distutils/sourcedist.html#specifying-the-files-to-distribute
    packages=[
        "spotdl",
        "spotdl.command_line",
        "spotdl.lyrics",
        "spotdl.lyrics.providers",
        "spotdl.encode",
        "spotdl.encode.encoders",
        "spotdl.metadata",
        "spotdl.metadata.embedders",
        "spotdl.metadata.providers",
        "spotdl.lyrics",
        "spotdl.lyrics.providers",
        "spotdl.authorize",
        "spotdl.authorize.services",
        "spotdl.helpers",
    ],
    version=__version__,
    install_requires=[
        "pathlib >= 1.0.1",
        "youtube_dl >= 2017.9.26",
        "pytube3 >= 9.5.5",
        "spotipy >= 2.12.0",
        "mutagen >= 1.41.1",
        "beautifulsoup4 >= 4.6.3",
        "unicode-slugify >= 0.1.3",
        "coloredlogs >= 14.0",
        "lyricwikia >= 0.1.8",
        "PyYAML >= 3.13",
        "appdirs >= 1.4.3",
        "tqdm >= 4.45.0"
    ],
    description="Download songs from YouTube using Spotify song URLs or playlists with albumart and meta-tags.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ritiek Malhotra",
    author_email="ritiekmalhotra123@gmail.com",
    license="MIT",
    python_requires=">=3.6",
    url="https://github.com/ritiek/spotify-downloader",
    download_url="https://pypi.org/project/spotdl/",
    keywords=[
        "spotify",
        "downloader",
        "download",
        "music",
        "youtube",
        "mp3",
        "album",
        "metadata",
    ],
    classifiers=[
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Multimedia",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Utilities",
    ],
    entry_points={"console_scripts": ["spotdl = spotdl:main"]},
)



from spotdl.metadata.provider_base import ProviderBase
from spotdl.metadata.provider_base import StreamsBase

from spotdl.metadata.exceptions import MetadataNotFoundError
from spotdl.metadata.exceptions import SpotifyMetadataNotFoundError
from spotdl.metadata.exceptions import YouTubeMetadataNotFoundError

from spotdl.metadata.embedder_base import EmbedderBase

from spotdl.metadata.formatter import format_string

